"""Remove error field from task_queue

Revision ID: 5686f64a46c8
Revises: dd35678aae90
Create Date: 2020-10-27 14:06:58.794701

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "5686f64a46c8"
down_revision = "dd35678aae90"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("task_queue", "error")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("task_queue", sa.Column("error", sa.VARCHAR(), autoincrement=False, nullable=True))
    # ### end Alembic commands ###
