"""Remove task_id from base_results table

Revision ID: dd35678aae90
Revises: 30fd8253d87f
Create Date: 2020-10-26 16:35:40.136241

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "dd35678aae90"
down_revision = "30fd8253d87f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("base_result", "task_id")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("base_result", sa.Column("task_id", sa.VARCHAR(), autoincrement=False, nullable=True))
    # ### end Alembic commands ###
